/*
 * Screen.c
 *
 *  Author: Погребняк Дмитрий
 */ 


#define F_CPU 8000000UL

#include "wg12232.h"
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include "font.h"
#include "Tools.h"

uint8_t encoder_pos = 0;
int8_t encoder_direction = 0;

const char str_logo[] PROGMEM = "WG12232A";
const char str_model[] PROGMEM = " YGH V";
const char str_author[] PROGMEM = "Автор:";
const char str_me[] PROGMEM = "Погребняк Д.Э.";
const char str_where[] PROGMEM = "Самара, 2013";

const char str_fly[] PROGMEM = "WG12232A-YGH-V#A";

typedef struct {
  int8_t x, y, z;
} Point3D;

typedef struct {
  int8_t x, y;
  uint8_t isVisible;
} Point3Dv;

typedef struct {
  uint8_t p1, p2;
} Line;

typedef struct {
  uint8_t cnt;
  uint8_t points[5];
} Surface;

typedef struct {
  int8_t x, y;
  uint8_t z;
  int8_t sx, sy;
} Star;

const Point3D points[] PROGMEM = {
  {-50, -50, -50},
  {50, -50, -50},
  {-50, 50, -50},
  {50, 50, -50},
  {-50, -50, 50},
  {-50, 50, 50},
  {50, 50, 50},
  {50, 0, 50},
  {50, -50, 0},
  {0, -50, 50},
};

const Line lines[] PROGMEM = {
  {0, 1}, {1, 3}, {3, 2}, {2, 0},
  {0, 4}, {4, 5}, {5, 2}, {5, 6}, {6, 3},
  {7, 8}, {8, 9}, {9, 7}, 
  {4, 9}, {1, 8}, {6, 7}
};


const Surface surfaces[] PROGMEM = {
  {4, {0, 2, 3, 1}},
  {4, {0, 4, 5, 2}},
  {4, {5, 6, 3, 2}},
  {5, {4, 0, 1, 8, 9}},
  {5, {6, 5, 4, 9, 7}},
  {5, {1, 3, 6, 7, 8}},
  {3, {9, 8, 7}},
};

const int16_t sintable[] PROGMEM = {
  0, 6, 13, 19, 25, 31, 38, 44, 50, 56, 62, 68, 74, 80, 86, 92, 98,
  104, 109, 115, 121, 126, 132, 137, 142, 147, 152, 157, 162, 167, 172, 177, 181,
  185, 190, 194, 198, 202, 206, 209, 213, 216, 220, 223, 226, 229, 231, 234, 237,
  239, 241, 243, 245, 247, 248, 250, 251, 252, 253, 254, 255, 255, 256, 256, 256
};

inline int16_t tsin(uint8_t period) {
  if (period >= 0x80) {
    return (period > 0xC0) ? -pgm_read_word(&sintable[0x100 - period]) : -pgm_read_word(&sintable[period - 0x80]);
  } else {
    return (period > 0x40) ? pgm_read_word(&sintable[0x80 - period]) : pgm_read_word(&sintable[period]);
  }
}

inline int16_t tcos(uint8_t period) {
  if (period >= 0x80) {
    return (period > 0xC0) ? pgm_read_word(&sintable[period - 0xC0]) : -pgm_read_word(&sintable[0xC0 - period]);
  } else {
    return (period > 0x40) ? -pgm_read_word(&sintable[period - 0x40]) : pgm_read_word(&sintable[0x40 - period]);
  }
}

Point3Dv vpoints[arraysize(points)];

inline void rotate_all_points(uint8_t xAngle, uint8_t yAngle, uint8_t x, uint8_t distAdd) {
  for (uint8_t i = 0; i < arraysize(points); i++) {
    int16_t ax = (int8_t)pgm_read_byte(&points[i].x);
    int16_t ay = (int8_t)pgm_read_byte(&points[i].y);
    int16_t az = (int8_t)pgm_read_byte(&points[i].z);
    int16_t s, c;
    s = tsin(xAngle);
    c = tcos(xAngle);
    int16_t t = (ay * c - az * s) / 256;
    az = (az * c + ay * s) / 256;
    ay = t;
    s = tsin(yAngle);
    c = tcos(yAngle);
    t = (ax * c - az * s) / 256;
    az = (az * c + ax * s) / 256;
    ax = t;
    az = az + 200 + distAdd;
    vpoints[i].x = ax * 34 / az + x;
    vpoints[i].y = ay * 34 / az + 16;
    vpoints[i].isVisible = 0;
  }
}

inline void find_visible_points() {
  for (uint8_t i = 0; i < arraysize(surfaces); i++) {
    uint8_t p1 = pgm_read_byte(&surfaces[i].points[0]);
    uint8_t p2 = pgm_read_byte(&surfaces[i].points[1]);
    uint8_t p3 = pgm_read_byte(&surfaces[i].points[2]);
    int8_t dx1 = vpoints[p1].x - vpoints[p2].x;
    int8_t dy1 = vpoints[p1].y - vpoints[p2].y;
    int8_t dx2 = vpoints[p3].x - vpoints[p2].x;
    int8_t dy2 = vpoints[p3].y - vpoints[p2].y;

    int16_t v = (int16_t)(dx1 * dy2) - (int16_t)(dy1 * dx2);
    if (v > 0) {
      vpoints[p1].isVisible = 1;
      vpoints[p2].isVisible = 1;
      vpoints[p3].isVisible = 1;
      uint8_t pcnt = pgm_read_byte(&surfaces[i].cnt);
      if (pcnt >= 4) 
        vpoints[pgm_read_byte(&surfaces[i].points[3])].isVisible = 1;
      if (pcnt >= 5)
        vpoints[pgm_read_byte(&surfaces[i].points[4])].isVisible = 1;

    }
  }
}

inline void draw_all_lines() {
  for (uint8_t i = 0; i < arraysize(lines); i++) {
    uint8_t p1 = pgm_read_byte(&lines[i].p1);
    uint8_t p2 = pgm_read_byte(&lines[i].p2);
    if (vpoints[p1].isVisible && vpoints[p2].isVisible)
      display_line(vpoints[p1].x, vpoints[p1].y, vpoints[p2].x, vpoints[p2].y, 0, 0xFF, 0, 0xFF);
  } 
}

uint32_t randomSeed = 123431245;

uint8_t random() {
  return (randomSeed = randomSeed * 0x08088405 + 1) >> 24;
}

int main(void)
{
  display_init();
  while(1)
  {
    text_set_font(FONT_BOLD);
    text_set_pos(32, 0);
    text_print_pgm(&str_logo, 0);

    text_set_font(FONT_NORMAL);
    text_print_pgm(&str_model, 0);

    text_set_font(FONT_SMALL);
    text_set_pos((DISPLAY_WIDTH - 32 - text_width_pgm(&str_author)) / 2 + 32, 10);
    text_print_pgm(&str_author, 0);

    text_set_font(FONT_NORMAL);
    text_set_pos((DISPLAY_WIDTH - 32 - text_width_pgm(&str_me)) / 2 + 32, 16);
    text_print_pgm(&str_me, 0);

    text_set_font(FONT_SMALL);
    text_set_pos((DISPLAY_WIDTH - 32 - text_width_pgm(&str_where)) / 2 + 32, 26);
    text_print_pgm(&str_where, 0);

    uint8_t xAngle1 = 123;
    uint8_t yAngle1 = 32;
    uint16_t distPeriod1 = 0;
    for (uint16_t f = 0; f < 500; f++) {
      xAngle1 += 3;
      yAngle1 += 1;
      distPeriod1 += 439;
      rotate_all_points(xAngle1, yAngle1, 16, ((tsin(distPeriod1 >> 8) + 280) * 100) >> 8);
      find_visible_points();
      display_fill_rect(0, 0, 32, 32, 0, 0, 0, 0);
      draw_all_lines();
      _delay_ms(50);
    }
    for (uint8_t page = 0; page < DISPLAY_PAGES; page++) {
      wg12232_command(WG12232_CMD_PAGE + page, WG12232_CHIP_BOTH);
      wg12232_command(WG12232_CMD_COLUMN + DISPLAY_START_COLUMN, WG12232_CHIP_BOTH);
      for (uint8_t x = 0; x < DISPLAY_HALF_WIDTH; x++) {
        wg12232_write(0xFF, WG12232_CHIP_BOTH);
        _delay_ms(20);
      }
    }
    for (uint8_t page = 0; page < DISPLAY_PAGES; page++) {
      wg12232_command(WG12232_CMD_PAGE + page, WG12232_CHIP_BOTH);
      wg12232_command(WG12232_CMD_COLUMN + DISPLAY_START_COLUMN, WG12232_CHIP_BOTH);
      for (uint8_t x = 0; x < DISPLAY_HALF_WIDTH; x++) {
        wg12232_write(0x00, WG12232_CHIP_BOTH);
        _delay_ms(5);
      }
    }

    text_set_font(FONT_BOLD);
    int16_t sx = DISPLAY_WIDTH - 1;
    void * startch = (void *)&str_fly;
    for(;;) {
      int16_t ox = sx;
      void * pch = startch;
      uint8_t ch = pgm_read_byte(pch++);
      if (!ch) 
        break;
      display_clear();
      do {
        uint8_t w =  text_width_char(ch);
        if (ox <= -w) {
          ox += w;
          sx = ox;
          startch = pch;
        } else {
          text_set_pos(ox, ((tsin(ox * 5 / 2) + 256) * 12) >> 8);
          text_print_char(ch, 0);
          ox += w;
        }
        if (ox >= DISPLAY_WIDTH)
          break;
        ch = pgm_read_byte(pch++);
      } while (ch);
      _delay_ms(80);
      sx--;
    }
    _delay_ms(1000);

    uint8_t xAngle2 = 12;
    uint8_t yAngle2 = 97;
    uint16_t distPeriod2 = 0;
    uint8_t xAngle3 = 231;
    uint8_t yAngle3 = 42;
    uint16_t distPeriod3 = 3276;
    for (uint16_t f = 0; f < 500; f++) {

      if (f > 128) {
        if (f <= 256) {
          wg12232_command(WG12232_CMD_START_LINE + ((f >> 2) & 31), WG12232_CHIP_LEFT);
        } else if (f <= 384) {
          wg12232_command(WG12232_CMD_START_LINE + ((-(f >> 2)) & 31), WG12232_CHIP_RIGHT);
        }
      }
      xAngle1 += 3;
      yAngle1 += 1;
      distPeriod1 += 439;
      rotate_all_points(xAngle1, yAngle1, 16, ((tsin(distPeriod1 >> 8) + 280) * 100) >> 8);
      find_visible_points();
      display_fill_rect(0, 0, 32, 32, 0, 0, 0, 0);
      draw_all_lines();

      xAngle2 += 2;
      yAngle2 += 3;
      distPeriod2 += 701;
      rotate_all_points(xAngle2, yAngle2, 61, ((tsin(distPeriod2 >> 8) + 280) * 100) >> 8);
      find_visible_points();
      display_fill_rect(45, 0, 32, 32, 0, 0, 0, 0);
      draw_all_lines();

      xAngle3 += 5;
      yAngle3 += 2;
      distPeriod3 += 307;
      rotate_all_points(xAngle3, yAngle3, 106, ((tsin(distPeriod3 >> 8) + 280) * 100) >> 8);
      find_visible_points();
      display_fill_rect(90, 0, 32, 32, 0, 0, 0, 0);
      draw_all_lines();

      _delay_ms(50);
    }
    wg12232_command(WG12232_CMD_START_LINE + 0, WG12232_CHIP_BOTH);

    display_clear();

    Star stars[64];

    for(uint8_t i = 0; i < arraysize(stars); i++) {
      stars[i].x = random();
      stars[i].y = random();
      stars[i].z = (random() >> 1) + 100;
      stars[i].sx = -1;
      stars[i].sy = -1;
    }

    uint8_t oscillate = 0;
    for (uint16_t f = 0; f < 500; f++) {
      int8_t osca = tsin(oscillate) / 8;
      int16_t so = tsin(osca);
      int16_t co = tcos(osca);
      oscillate++;
      for(uint8_t i = 0; i < arraysize(stars); i++) {
        int16_t dx = stars[i].x * co + stars[i].y * so;
        int16_t dy = stars[i].y * co - stars[i].x * so;
        int16_t div = stars[i].z * 4;
        dx /= div;
        dy /= div;
        if ((dx <= -128) || (dx >= 128) || (dy <= -128) || (dy >= 128)) {
          stars[i].sx = -1;
          stars[i].sy = -1;
          stars[i].z  = 0;
        } else {
          dx += DISPLAY_WIDTH / 2;
          dy += DISPLAY_HEIGHT / 2;
          if ((dx < 0) || (dx >= DISPLAY_WIDTH))
            dx = -1;
          if ((dy < 0) || (dy >= DISPLAY_WIDTH))
            dy = -1;
          stars[i].sx = dx;
          stars[i].sy = dy;
          stars[i].z -= 16;
        }
        if (stars[i].z < 20) {
          stars[i].x = random();
          stars[i].y = random();
          stars[i].z = (random() >> 2) + 190;
        }
      }
      display_fill(0xFF);
      for(uint8_t i = 0; i < arraysize(stars); i++) {
        if ((stars[i].sx >= 0) && (stars[i].sy >= 0))
          display_pixel(stars[i].sx, stars[i].sy, 0x00, 0x00);
      }

      _delay_ms(80);
    }

    display_clear();

    for (uint8_t f = 0; f < 3; f++) {
      text_set_font(f);
      display_clear();
      text_set_pos(0, 0);
      for (uint8_t c = 24; c <= 194; c++) {
        text_print_char(c, 0);
        _delay_ms(10);
        if (text_get_X() >= DISPLAY_WIDTH - 4) {
          uint8_t th = text_height();
          uint8_t ny = text_get_Y() + th;
          if (ny > (DISPLAY_HEIGHT - th)) {
            ny = 0;
            _delay_ms(2000);
            display_clear();
          }
          text_set_pos(0, ny);
        }
      }
      _delay_ms(4000);
    }

    display_clear();
  }
}